package io.scanbot.sdk.flutter.camera

import android.content.Context
import android.graphics.Color
import android.graphics.PointF
import android.view.View
import android.widget.FrameLayout
import com.google.gson.FieldNamingPolicy
import com.google.gson.GsonBuilder
import io.flutter.plugin.common.BinaryMessenger
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.platform.PlatformView
import io.scanbot.sdk.ScanbotSDK
import io.scanbot.sdk.flutter.*
import io.scanbot.sdk.flutter.camera.config.CroppingResult
import io.scanbot.sdk.flutter.utils.convertDpToPx
import io.scanbot.sdk.flutter.utils.throttleFirst
import io.scanbot.sdk.persistence.Page
import io.scanbot.sdk.persistence.PageFileStorage
import io.scanbot.sdk.ui.EditPolygonImageView
import io.scanbot.sdk.ui.MagnifierView
import io.scanbot.sdk.ui.view.interactor.DetectPolygonOnPageUseCase
import io.scanbot.sdk.util.PolygonHelper
import kotlinx.coroutines.*
import kotlinx.coroutines.flow.MutableSharedFlow
import kotlinx.coroutines.flow.catch
import kotlinx.coroutines.flow.launchIn
import kotlinx.coroutines.flow.onEach
import kotlinx.coroutines.flow.single
import kotlin.coroutines.CoroutineContext


class ScanbotCroppingPlatformView internal constructor(
    private val context: Context,
    private val messenger: BinaryMessenger,
    private val id: Int,
    private val creationParams: Json?
) : PlatformView,
    MethodChannel.MethodCallHandler, CoroutineScope {

    private val rootView: FrameLayout = FrameLayout(context)
    private val editPolygonImageView = EditPolygonImageView(context, null)
    private val magnifierView = MagnifierView(context, null)
    private val scanbotSDK = ScanbotSDK(context)
    private val contourDetector = scanbotSDK.createContourDetector()
    private val pageFileStorage = scanbotSDK.createPageFileStorage()
    private var croppingViewChannel = MethodChannel(messenger, "scanbot_sdk_cropping_$id")
    private var page: Page? = null
    private val returnContext = Dispatchers.Main
    private val actions: MutableSharedFlow<EditPolyAction> = MutableSharedFlow()
    val gson = GsonBuilder()
        .setFieldNamingPolicy(FieldNamingPolicy.IDENTITY)
        .registerTypeAdapter(ByteArray::class.java, ByteArrayToBase64TypeAdapter())
        .create()

    private var job: Job = Job()
    override val coroutineContext: CoroutineContext
        get() = Dispatchers.Default + job

    override fun getView(): View {
        return rootView
    }

    init {
        initView()
    }

    private fun initView() {
        rootView.addView(
            editPolygonImageView,
            FrameLayout.LayoutParams(
                FrameLayout.LayoutParams.MATCH_PARENT,
                FrameLayout.LayoutParams.MATCH_PARENT
            )
        )
        rootView.addView(
            magnifierView,
            FrameLayout.LayoutParams(
                FrameLayout.LayoutParams.MATCH_PARENT,
                FrameLayout.LayoutParams.MATCH_PARENT
            )
        )

        if (!scanbotSDK.licenseInfo.isValid) {
            launch(returnContext) {
                val json = scanbotSDK.getLicenseStatusJson()
                croppingViewChannel.invokeMethod(
                    "onLicenseError",
                    json.toString()
                )
            }
        }

        creationParams?.checkParam<Json>("page") {
            launch {
                val newPage = it.toPage(gson)
                page = newPage
                processPage(newPage)
            }
        }
        applyAppearanceConfig()
        actions.throttleFirst(300).onEach { action ->
            when (action) {
                is EditPolyAction.Detect -> {
                    detectAndSetPolygon(action.result)
                }

                is EditPolyAction.Reset -> {
                    withContext(Dispatchers.Main) {
                        editPolygonImageView.polygon = PolygonHelper.getFullPolygon()
                        action.result.success("success")
                    }
                }

                is EditPolyAction.RotateCCW -> {
                    withContext(Dispatchers.Main) {
                        editPolygonImageView.rotateCounterClockwise()
                        action.result.success("success")
                    }
                }

                is EditPolyAction.RotateCW -> {
                    withContext(Dispatchers.Main) {
                        editPolygonImageView.rotateClockwise()
                        action.result.success("success")
                    }
                }

                is EditPolyAction.GetPolygon -> {
                    val rotation = editPolygonImageView.rotation
                    val times = rotation / (-90);
                    val polygon = editPolygonImageView.polygon
                    withContext(Dispatchers.Main) {
                        action.result.success(getJsonForPolygonData(polygon, times))
                    }
                }

                is EditPolyAction.NewPage -> {
                    val newPage = (action.methodCall.arguments as Json).toPage(gson)
                    processPage(newPage)
                    action.result.success("success")
                }
            }
        }.launchIn(this)
        croppingViewChannel.setMethodCallHandler(this)
    }

    private suspend fun processPage(page: Page) {
        val bitmap =
            pageFileStorage.getPreviewImage(page.pageId, PageFileStorage.PageFileType.ORIGINAL)

        bitmap?.let { image ->
            withContext(Dispatchers.Main) {
                editPolygonImageView.setImageBitmap(image)
                editPolygonImageView.setMagnifier(magnifierView)
                magnifierView.setupMagnifier(editPolygonImageView)

                DetectPolygonOnPageUseCase(
                    contourDetector,
                    pageFileStorage
                ).detectPolygonOnPage(
                    page
                ).onEach { result ->
                    withContext(Dispatchers.Main) {
                        editPolygonImageView.setLines(
                            result?.horizontal ?: emptyList(),
                            result?.vertical ?: emptyList()
                        )
                        if (page.polygon.isEmpty() || page.polygon == PolygonHelper.getFullPolygon()) {
                            editPolygonImageView.polygon =
                                result?.polygon?.takeIf { it.isNotEmpty() }
                                    ?: PolygonHelper.getFullPolygon()
                        } else {
                            editPolygonImageView.polygon = page.polygon
                        }
                        editPolygonImageView.setRotationAnimated(0f)
                    }
                }.single()
            }

        }
    }

    private fun applyAppearanceConfig() {
        creationParams?.toCroppingViewConfig(gson)?.let { config ->
            config.edgeLineWidth?.let {
                editPolygonImageView.setEdgeWidth(convertDpToPx(context, it))
            }
            config.edgeColor?.let {
                editPolygonImageView.setEdgeColor(Color.parseColor(it))
            }
            config.anchorPointsColor?.let {
                editPolygonImageView.setAnchorPointsColor(Color.parseColor(it))
            }

            config.edgeColorOnLine?.let {
                editPolygonImageView.setEdgeColorOnLine(Color.parseColor(it))
            }

            config.borderInsets?.let { inset ->
                editPolygonImageView.setPadding(
                    inset.left?.let { value -> convertDpToPx(context, value).toInt() } ?: 0,
                    inset.top?.let { value -> convertDpToPx(context, value).toInt() } ?: 0,
                    inset.right?.let { value -> convertDpToPx(context, value).toInt() } ?: 0,
                    inset.bottom?.let { value -> convertDpToPx(context, value).toInt() } ?: 0,
                )
            }
        }
    }


    override fun onMethodCall(methodCall: MethodCall, result: MethodChannel.Result) {
        launch {
            when (methodCall.method) {
                "rotateCW" -> {
                    actions.emit(EditPolyAction.RotateCW(result))
                }

                "rotateCCW" -> {
                    actions.emit(EditPolyAction.RotateCCW(result))
                }

                "reset" -> {
                    actions.emit(EditPolyAction.Reset(result))
                }

                "getEditedPolygon" -> {
                    actions.emit(EditPolyAction.GetPolygon(result))
                }

                "detect" -> {
                    actions.emit(EditPolyAction.Detect(result))
                }

                "setNewPage" -> {
                    actions.emit(EditPolyAction.NewPage(methodCall, result))
                }

                "dispose" -> {
                    dispose()
                    result.success("success")
                }

                else -> result.notImplemented()
            }
        }
    }

    private suspend fun CoroutineScope.detectAndSetPolygon(result: MethodChannel.Result) {
        page?.let {
            DetectPolygonOnPageUseCase(
                contourDetector,
                pageFileStorage
            ).detectPolygonOnPage(
                it
            ).onEach { data ->
                withContext(Dispatchers.Main) {
                    croppingViewChannel.invokeMethod("onShowHeavyProcessing", true)
                }
                val detectedPolygon =
                    data?.polygon ?: page?.polygon ?: PolygonHelper.getFullPolygon()
                editPolygonImageView.polygon = detectedPolygon
                editPolygonImageView.setLines(
                    data?.horizontal ?: emptyList(),
                    data?.vertical ?: emptyList()
                )
                result.success("success")
                withContext(Dispatchers.Main) {
                    croppingViewChannel.invokeMethod("onShowHeavyProcessing", false)
                }

            }.catch {
                result.success("error: ${it.localizedMessage}")
                withContext(Dispatchers.Main) {
                    croppingViewChannel.invokeMethod("onShowHeavyProcessing", true)
                }

            }.single()
        }
    }

    private fun getJsonForPolygonData(
        polygon: List<PointF>,
        rotation: Float
    ): String {
        return gson.toJson(CroppingResult(polygon, rotation.toInt()))
    }


    override fun dispose() {
        job.cancel()
    }

    override fun onFlutterViewDetached() {
        super.onFlutterViewDetached()
    }

    override fun onFlutterViewAttached(flutterView: View) {
        super.onFlutterViewAttached(flutterView)
    }

}

sealed class EditPolyAction(val result: MethodChannel.Result) {
    class NewPage(val methodCall: MethodCall, result: MethodChannel.Result) : EditPolyAction(result)
    class GetPolygon(result: MethodChannel.Result) : EditPolyAction(result)
    class Reset(result: MethodChannel.Result) : EditPolyAction(result)
    class Detect(result: MethodChannel.Result) : EditPolyAction(result)
    class RotateCW(result: MethodChannel.Result) : EditPolyAction(result)
    class RotateCCW(result: MethodChannel.Result) : EditPolyAction(result)
}